#  """
#    Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import shutil
from os import popen
from os.path import join, isdir
from sys import platform
from enum import Enum


class WiliotPackages(Enum):
    CLOUD = ['pywiliot-api']
    CORE = ['pywiliot-core', 'pywiliot-api']
    TOOLS = ['pywiliot-tools', 'pywiliot-core', 'pywiliot-api']
    DEPLOYMENT = ['pywiliot-deployment-tools', 'pywiliot-tools', 'pywiliot-core', 'pywiliot-api']
    TESTERS = ['pywiliot-testers', 'pywiliot-tools', 'pywiliot-core', 'pywiliot-api']


def update_internal_wiliot_packages(wiliot_repo=WiliotPackages.CLOUD, branch_name='master',
                                    overwrite=False, no_wiliot_dep=False,
                                    username_bitbucket=None, password_bitbucket=None,
                                    install_from_local_folder='', branch_name_for_dep='master'):
    # check platform command prefix
    python_ver = ''
    if platform == "darwin" or platform == 'linux':
        # macos or linux
        python_ver = '3'
    # install pywiliot and requirements:
    # update dependencies
    print("----------------------------------\nupdate pip:\n----------------------------------")
    p = popen(f'python{python_ver} -m pip install --upgrade pip')
    rsp = p.read()
    print(rsp)
    pkg_versions = []
    req_status = []
    for repo_num, repo_name in enumerate(wiliot_repo.value):
        package_name = repo_name.replace('py', '')
        folder_name = package_name.replace('-', '_')
        print('check if package already exist')
        p = popen(f'pip{python_ver} show {package_name}')
        rsp = p.read()
        print(rsp)
        if package_name in rsp:
            if overwrite:
                print(f'--------------------------------------------------------------------\n'
                      f'uninstall existing package: {package_name}\n'
                      f'--------------------------------------------------------------------\n')
                p = popen(f'pip{python_ver} uninstall {package_name} -y')
            else:
                print(f'--------------------------------------------------------------------\n'
                      f'for uninstall existing package: {package_name}, please press ENTER\n'
                      f'--------------------------------------------------------------------\n')
                p = popen(f'pip{python_ver} uninstall {package_name}')
            rsp = p.read()
            print(rsp)
        print(f'--------------------------------------------------------------------\n'
              f'update wiliot package: {package_name}\n'
              f'--------------------------------------------------------------------')
        if install_from_local_folder:
            local_repo_path = join(install_from_local_folder, repo_name)
            if not isdir(local_repo_path):
                print(f'{repo_name} does not exist in local folder, skip installation')
                pkg_versions.append('')
                req_status.append(False)
                continue
            p = popen(f'pip{python_ver} install -e {local_repo_path}')
            rsp = p.read()
            print(rsp)
        elif username_bitbucket is None and password_bitbucket is None:
            if repo_num == 0:
                p = popen(f'pip{python_ver} install git+ssh://git@bitbucket.org/wiliot/{repo_name}.git'
                          f'@{branch_name}#egg={folder_name} --upgrade')
            else:
                p = popen(f'pip{python_ver} install git+ssh://git@bitbucket.org/wiliot/{repo_name}.git'
                          f'@{branch_name_for_dep}#egg={folder_name} --upgrade')
            rsp = p.read()
            print(rsp)
            if 'Successfully installed' not in rsp.split('\n')[-2]:
                print(f'problem with ssh key credentials. Please rerun the script using username (-u) and password (-p).'
                      f'\nTo extract Attlasian username and password go to '
                      f'https://wiliot.atlassian.net/wiki/spaces/SW/pages/2890662124/'
                      f'PyWiliot+Installation+Guidelines+-+Private#using-username-and-password%3A')
                return
        elif username_bitbucket is not None and password_bitbucket is not None:
            if repo_num == 0:
                p = popen(f'pip{python_ver} install git+https://{username_bitbucket}:{password_bitbucket}@bitbucket.org'
                          f'/wiliot/{repo_name}.git@{branch_name}#egg={folder_name} --upgrade')
            else:
                p = popen(f'pip{python_ver} install git+https://{username_bitbucket}:{password_bitbucket}@bitbucket.org'
                          f'/wiliot/{repo_name}.git#egg={folder_name} --upgrade')
            rsp = p.read()
            print(rsp)
        else:
            print('you Must specified both username and password or do NOT specified neither '
                  'for installation using SSH key')
            return

        print(f'check wiliot package: {package_name}')
        p = popen(f'pip{python_ver} show {package_name}')
        rsp = p.read()
        print(rsp)
        if 'Location' in rsp:
            pkg_versions.append(rsp.split('Version: ')[1].split('\n')[0])
            req_path = rsp.split('Location: ')[-1].split('\n')[0]
            if install_from_local_folder:
                site_packages_path = join(req_path, folder_name)
                if isdir(site_packages_path):
                    shutil.rmtree(site_packages_path)
                shutil.copytree(join(install_from_local_folder, repo_name, folder_name), site_packages_path)
                req_path = join(install_from_local_folder, repo_name)

            print("--------------------------------------------------------------------\n"
                  "update dependencies:\n"
                  "--------------------------------------------------------------------")
            requirements_path = join(req_path, folder_name, 'requirements.txt')
            p = popen(f'pip{python_ver} install -r {requirements_path}')
            rsp = p.read()
            print(rsp)
            req_status.append('ERROR' not in rsp)
        else:
            pkg_versions.append('')
            req_status.append(False)

        if no_wiliot_dep:
            break

    print('done installation')
    print("************************************************************************\n"
          "********************            Summary            *********************\n"
          "************************************************************************")
    for i, repo_name in enumerate(wiliot_repo.value):
        package_name = repo_name.replace('py', '')
        is_installed = f'installed [{pkg_versions[i]}]' if pkg_versions[i] else 'NOT installed'
        sum_str = f'{package_name} was {is_installed}'
        if is_installed:
            req_is_installed = 'installed' if req_status[i] else 'NOT installed'
            sum_str += f'and the dependencies were {req_is_installed}'
        print(sum_str)
        if no_wiliot_dep:
            break


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='Update Wiliot Internal Packages')
    parser.add_argument('-p', '--package_name', help='The package name. valid option(wiliot-api, wiliot-core, '
                                                     'wiliot-tools, wiliot-testers, wiliot-deployment-tools')
    parser.add_argument('-b', '--branch_name', help='The branch name you want to use. the default is master',
                        default='master')
    parser.add_argument('-w', '--overwrite', help='If specified, all wiliot existed packages will be overwrite '
                                                  'with the new installation', default=False, action='store_true')
    parser.add_argument('-n', '--no_wiliot_dep', help='If specified, install the main package without wiliot '
                                                      'dependencies packages', default=False, action='store_true')
    parser.add_argument('-u', '--username', help='If specified, install based username and password and not SHH key',
                        default=None)
    parser.add_argument('-pass', '--password', help='If specified, install based username and password and not SHH key',
                        default=None)
    parser.add_argument('-l', '--local_folder', help='If specified, install packages from the specified local folder')
    parser.add_argument('-b-for-dep', '--branch_for_dependencies',
                        help='if specified, all wiliot package dependencies will be installed for this branch '
                             '(default master)', default='master')
    args = parser.parse_args()
    package_name = args.package_name
    branch_name = args.branch_name
    is_overwrite = args.overwrite
    is_no_wiliot_dep = args.no_wiliot_dep
    username = args.username
    password = args.password
    local_folder = args.local_folder
    branch_for_dep = args.branch_for_dependencies
    err_msg = 'please add package name, e.g. update_wiliot_packages.py -p wiliot-api.\n' \
              'The valid packages names are: wiliot-api, wiliot-core, wiliot-tools, wiliot-testers, ' \
              'wiliot-deployment-tools'
    try:
        repo_name = None
        for package in WiliotPackages.__members__.values():
            if package_name in package.value[0]:
                repo_name = package
                break
        if repo_name is None:
            print(err_msg)
        else:
            update_internal_wiliot_packages(wiliot_repo=repo_name, branch_name=branch_name,
                                            overwrite=is_overwrite, no_wiliot_dep=is_no_wiliot_dep,
                                            username_bitbucket=username, password_bitbucket=password,
                                            install_from_local_folder=local_folder, branch_name_for_dep=branch_for_dep)
    except Exception as e:
        print('Error during update packages  [ERROR: {}]'.format(e))



#  """
#    Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import logging

import numpy as np
import pandas as pd
import copy
import re
from enum import Enum
import os

packet_data_dict = {
    'adv_address': (0, 6),
    'en': (6, 1),
    'type': (7, 1),
    'data_uid': (8, 2),
    # 'group_id': (10, 3), #group id is only 22bits, we have a dedicated function for it
    'nonce': (13, 4),
    'enc_uid': (17, 6),
    'mic': (23, 6),
    'enc_payload': (29, 8),
}
gw_attributes = {
    'gw_packet': 'gw_packet',
    'rssi': 'rssi',
    'stat_param': 'stat_param',
    'time_from_start': 'time_from_start',
    'counter_tag': 'counter_tag',
    'is_valid_tag_packet': 'is_valid_tag_packet',
}
general_data = {
    'gw_process': 'gw_process',
    'is_valid_packet': 'is_valid_packet'
}
packet_length = 78
packet_tag_length = 74
rssi_char_last_index = 76
stat_param_last_index = 80

max_stat_param_val = 65535  # 2 bytes


class InlayTypes(Enum):
    TEO_086 = '086'
    TIKI_096 = '096'
    TIKI_099 = '099'
    BATTERY_107 = '107'
    TIKI_117 = '117'
    TIKI_118 = '118'
    TIKI_121 = '121'
    TIKI_122 = '122'


class Packet(object):
    """
    Wiliot Packet Object

        :param raw_packet: the raw packet to create a Packet object
        :type raw_packet: str or dict

        :return:
    """

    def __init__(self, raw_packet, time_from_start=None, custom_data=None,
                 inlay_type=None, logger_name=None):
        """
        :param raw_packet:
        :type raw_packet: str or dict
        :param logger_name: the log file we want to write into it.
        :type time_from_start: str
        :param time_from_start: the time when the packet was received according to the gw timer
        :type time_from_start: float
        :param custom_data: the packet custom data as dictionary
        :type custom_data: dict
        :param inlay_type: the antenna/ inlay type (e.g. tiki, teo)
        :type inlay_type: str
        """

        self.is_valid_packet = False
        if logger_name is not None:
            self.logger = logging.getLogger(logger_name)
        else:
            self.logger = None

        if type(raw_packet) is dict:
            if raw_packet['is_valid_tag_packet']:
                get_packet_content = self.get_packet_content(raw_packet['packet'], get_gw_data=True)
                self.packet_data = {
                    'raw_packet': self.get_packet_content(raw_packet['packet']),
                    'adv_address': raw_packet['adv_address'],
                    'decrypted_packet_type': (int(raw_packet['packet'][24:25], 16) & 0xc) >> 2,
                    'group_id': self.extract_group_id(raw_packet['group_id'])
                }

                self.gw_data = {
                    'gw_packet': np.array(get_packet_content[packet_tag_length:]),
                    'rssi': np.array(raw_packet['rssi']),
                    'stat_param': np.array(raw_packet['stat_param']),  # gw_time
                    'time_from_start': np.array(raw_packet['time_from_start']),  # pc_time
                    'counter_tag': np.array(raw_packet['counter_tag']),
                    'is_valid_tag_packet': np.array(raw_packet['is_valid_tag_packet']),
                }
                self.gw_process = True
                self.is_valid_packet = True
            else:
                self.is_valid_packet = False
        elif type(raw_packet) is str:
            try:
                if not any(ext in raw_packet for ext in ['user_event', 'Command Complete Event']):
                    packet_data = self.get_packet_content(raw_packet)
                    raw_packet = self.get_packet_content(raw_packet, get_gw_data=True)
                    rssi_hex = raw_packet[packet_tag_length:rssi_char_last_index]
                    stat_param_hex = raw_packet[rssi_char_last_index:stat_param_last_index]

                    self.packet_data = {
                        'raw_packet': packet_data,
                        'adv_address': packet_data[:12],
                        'decrypted_packet_type': (int(packet_data[24:25], 16) & 0xc) >> 2,
                        'group_id': self.extract_group_id(packet_data[20:26]),
                    }

                    rssi_in = np.array(float('nan'))
                    stat_param_in = np.array(float('nan'))
                    if rssi_hex != '':
                        rssi_in = np.array(int(rssi_hex, base=16))
                    if stat_param_hex != '':
                        stat_param_in = np.array(int(stat_param_hex, base=16))
                    self.gw_data = {
                        'gw_packet': np.array(raw_packet[packet_tag_length:]),
                        'rssi': rssi_in,
                        'stat_param': stat_param_in,
                        'time_from_start': np.array(time_from_start),
                        'counter_tag': np.array(float('nan')),
                        'is_valid_tag_packet': np.array(float('nan')),
                    }

                    self.gw_process = False
                    self.is_valid_packet = True
                else:
                    self.is_valid_packet = False
            except Exception as e:
                self.printing_and_logging('got corrupted packet: {}'.format(raw_packet))

        if not self.is_valid_packet:
            self.printing_and_logging('Invalid Packet: {}'.format(raw_packet))
        self.decoded_data = {}
        if custom_data is not None:
            self.custom_data = custom_data
        else:
            self.custom_data = {}
        self.inlay_type = inlay_type
        if self.is_valid_packet:

            self.packet_data['flow_ver'] = hex(
                int(self.packet_data['adv_address'][0:2] + self.packet_data['adv_address'][-2:], 16))

            self.packet_data['test_mode'] = 0
            flow_version = hex(int(self.packet_data['flow_ver'], 16))

            for key in packet_data_dict.keys():
                packet_data_value = self.get_attribute(self.packet_data['raw_packet'], packet_data_dict.get(key))
                self.packet_data[key] = packet_data_value

            if flow_version < hex(0x42c):
                if 'FFFFFFFF' in self.packet_data['adv_address']:
                    self.packet_data['test_mode'] = 1
            elif flow_version < hex(0x500):
                if self.packet_data['adv_address'].startswith('FFFF') or self.packet_data['adv_address'].endswith(
                        'FFFF'):
                    self.packet_data['test_mode'] = 1
            else:
                if int(self.packet_data['data_uid'], 16) == 5:
                    self.packet_data['test_mode'] = 1
                    # TX LPM\HPM:
                    tx_lpm_hpm = self.packet_data['decrypted_packet_type'] & 1
                    self.decoded_data['tx_lpm_hpm'] = tx_lpm_hpm
                    if tx_lpm_hpm == 0:
                        self.decoded_data['tx_lpm_hpm_str'] = 'LPM'
                    else:
                        self.decoded_data['tx_lpm_hpm_str'] = 'HPM'
                    # battery mode:
                    battery_tag_ind = (self.packet_data['decrypted_packet_type'] & 2) >> 1
                    self.decoded_data['battery_tag_ind'] = battery_tag_ind

    def __len__(self):
        """
        gets number of sprinkler occurrences in packet
        """
        return self.gw_data['rssi'].size

    def __eq__(self, packet):
        """
        packet comparison
        """
        if self.is_same_sprinkler(packet):
            if packet.gw_data['gw_packet'].item() == self.gw_data['gw_packet'].item():
                return True
        return False

    def __str__(self):
        """
        packet print method
        """
        return str(
            'packet_data={packet_data}, gw_data={gw_data}'.format(packet_data=self.packet_data, gw_data=self.gw_data))

    def printing_and_logging(self, message):
        if self.logger:
            self.logger.info(message)
        else:
            print(message)

    def is_packet_from_bridge(self):
        return self.packet_data['raw_packet'][0] != '0'

    def get_payload(self):
        return self.packet_data['raw_packet'][16:74]

    def extract_group_id(self, raw_group_id, zero_bits=None, num_char=6):
        """
        extract group id from the raw group id packet
        :param raw_group_id: as it received by the gw
        :type raw_group_id: str
        :param zero_bits: number of bits from the group id that were taken to other feauture an should be ignored
        :type zero_bits: list
        :param num_char: number of group id charachters in the packets
        :type num_char: int
        :return: the group id
        :rtype: str
        """
        # need to zero 2 bits since 2 MSB bits of the group id are for decrypted_packet_type:
        # if raw group id is: 123456 --> the actual data is = 563412 --> the bit-location:{5: [16:20], 6:[20:24],
        # 3:[8:12], 4:[12:16], 1:[0:4], 2:[4:8]}
        if zero_bits is None:
            zero_bits = [16, 17]
        b = self.hex2bin(raw_group_id)
        if zero_bits is not None:
            b_list = list(b)
            for z in zero_bits:
                b_list[z] = '0'
            b = ''.join(b_list)
        return hex(int(b, 2))[2:].zfill(num_char)

    @staticmethod
    def hex2bin(hex, min_dig=0, zfill=True):
        b = bin(int(hex, 16))[2:]
        if zfill:
            b = bin(int(hex, 16))[2:].zfill(24)
        if len(b) > min_dig:
            pass
        else:
            b = b.zfill(min_dig)
        return b

    def set_inlay_type(self, inlay_type):
        self.inlay_type = inlay_type

    def is_in(self, packet):
        """
        is packet contains another packet

        :param packet: the other packet to verify
        :type packet: Packet

        :return: bool
        """
        if self.is_same_sprinkler(packet):
            if packet.gw_data['gw_packet'].item() in self.gw_data['gw_packet']:
                return True
        return False

    def get_packet(self):
        """
        gets raw packet string
        """
        return str(self.packet_data['raw_packet'])

    def split_packet(self, index):
        """
        split packet by index
        """
        packet_a = self.copy()
        packet_b = self.copy()
        remain = len(self) - index
        for key in self.gw_data.keys():
            for i in range(index):
                packet_a.gw_data[key] = np.delete(packet_a.gw_data[key], -1)

            for i in range(remain):
                packet_b.gw_data[key] = np.delete(packet_b.gw_data[key], 0)

        return packet_a, packet_b

    def copy(self):
        return copy.deepcopy(self)

    def sort(self):
        """
        sort gw_data lists according to gw_time
        """
        isort = np.argsort(self.gw_data['stat_param'])
        for key in self.gw_data.keys():
            self.gw_data[key] = self.gw_data[key][isort]

    def get_average_rssi(self):
        return np.average(self.gw_data['rssi'])

    # @staticmethod
    def is_short_packet(self):
        return len(self.get_packet_string(process_packet=False)) < packet_length

    @staticmethod
    def get_packet_content(raw_packet, get_gw_data=False):
        if 'process_packet' in raw_packet:
            raw_packet = raw_packet.split('(')[1]
            raw_packet = re.sub(r'\W+', '', raw_packet)

        if get_gw_data:
            return raw_packet
        else:
            return raw_packet[0:packet_tag_length]

    def get_packet_string(self, i=0, gw_data=True, process_packet=True):
        """
        gets process_packet string
        """
        process_packet_string = ['', '']
        if process_packet:
            process_packet_string = ['process_packet("', '")']

        if not self.is_valid_packet:
            return None
        if gw_data:
            gw_data = self.gw_data['gw_packet'].take(i)
        else:
            gw_data = ''
        return '{raw_packet}{gw_data}'.format(raw_packet=self.packet_data['raw_packet'],
                                              gw_data=gw_data).join(process_packet_string)

    def get_adva(self):
        return self.packet_data['adv_address']

    def get_flow(self):
        flow_version = hex(int(self.packet_data['flow_ver'], 16))
        return flow_version

    def get_rssi(self):
        return str(self.gw_data['rssi'])

    def get_packet_data_names(self):
        """
        extract all keys name which can potentially be a column in packet df
        :return:
        :rtype: list
        """
        dict_temp = self.__dict__
        keys = list(dict_temp.keys())
        for k in dict_temp.keys():
            if isinstance(dict_temp[k], dict):
                keys = keys + list(dict_temp[k].keys())
        return keys

    @staticmethod
    def get_attribute(raw_packet, loc_length):
        loc = loc_length[0] * 2
        length = loc_length[1] * 2
        return raw_packet[loc:loc + length]

    def is_same_sprinkler(self, packet):
        raw_packet = packet.packet_data['raw_packet']
        if raw_packet == self.packet_data['raw_packet']:
            return True
        return False

    def append_to_sprinkler(self, packet, output_log=True):
        status = True
        if self.is_same_sprinkler(packet):
            for i in range(len(packet)):
                try:
                    # stat param should be unique (time + rssi for same packet)- we make sure no duplications are added.
                    if np.take(packet.gw_data['stat_param'], i).item() not in self.gw_data['stat_param'] or \
                            np.take(packet.gw_data['time_from_start'], i).item() not in self.gw_data['time_from_start']:
                        for key in gw_attributes.keys():
                            self.gw_data[key] = np.append(self.gw_data[key], np.take(packet.gw_data[key], i))
                        for key in self.custom_data.keys():
                            if key in packet.custom_data.keys():
                                self.custom_data[key] = np.append(self.custom_data[key],
                                                                  np.take(packet.custom_data[key], i))
                            else:
                                self.custom_data[key] = np.append(self.custom_data[key], [None])
                    else:
                        if output_log:
                            self.printing_and_logging('Tried to add duplicated packet to sprinkler {}'.format(packet))
                except Exception as e:
                    self.printing_and_logging('Failed to add packet {} to sprinkler, exception: {}'
                                              .format(packet, str(e)))
        else:
            self.printing_and_logging('Not from the same sprinkler')
            status = False
        # self.sort()
        return status

    def as_dict(self, sprinkler_index=None):  # None not tested
        packet_data = self.packet_data.copy()
        sprinkler_gw_data = self.gw_data.copy()
        custom_data = self.custom_data.copy()
        if sprinkler_index is not None:
            if sprinkler_index > self.gw_data['stat_param'].size:
                return None
            for gw_attr in gw_attributes.keys():
                sprinkler_gw_data[gw_attr] = np.take(self.gw_data[gw_attr], sprinkler_index)
            for custom_attr in self.custom_data.keys():
                custom_data[custom_attr] = np.take(self.custom_data[custom_attr], sprinkler_index)
            dict_len = 1
        else:
            dict_len = len(self)
            for k, v in packet_data.items():
                packet_data[k] = [v] * dict_len

        data = {**packet_data, **sprinkler_gw_data, **custom_data}
        data['gw_process'] = [self.gw_process] * dict_len  # bool should be a list for pd.DataFrame.from_dict(data)
        data['is_valid_packet'] = [self.is_valid_packet] * dict_len  # bool should be a list
        data['inlay_type'] = [self.inlay_type] * dict_len
        return data

    def as_dataframe(self, sprinkler_index=None):
        data = self.as_dict(sprinkler_index=sprinkler_index)
        packet_df = pd.DataFrame.from_dict(data)

        return packet_df

    def get_per(self, expected_sprinkler_count=6):
        """
        Calculates the packet per at the sprinkler
        @param expected_sprinkler_count - in case of no beacons environment, sprinkler can be bigger than 6.
        @return packet per at percentage
        """
        return 100 * (1 - len(self) / expected_sprinkler_count)

    def get_tbp(self):
        """
        calculates the rate of packets from the same sprinkler
        :return: min_times_found - in msec
        :rtype: int
        """

        def triad_ratio_logic(diff_time_1, diff_time_2, ratio=1.0, error=10):
            """ estimate the time between successive packet according to only 3 packets out of 6 """
            if abs(diff_time_1 - ratio * diff_time_2) <= diff_time_2 / error:
                return True
            elif abs(diff_time_1 - (1 / ratio) * diff_time_2) <= diff_time_1 / error:
                return True
            else:
                return False

        def estimate_diff_packet_time(times_list, pc_time_list):
            if times_list.size < 3:
                return None
            sort_idx = pc_time_list.argsort()
            pc_time_list_sorted = pc_time_list.copy()[sort_idx]
            times_list_sorted = times_list.copy()[sort_idx]

            fix_sort_idx = []
            ind = 0
            pc_time_unique, pc_time_counts = np.unique(pc_time_list_sorted, return_counts=True)
            for unique, counts in zip(pc_time_unique, pc_time_counts):
                if counts > 1:  # several packets
                    ind_with_same_time = [i for i, t in enumerate(pc_time_list_sorted) if t == unique]
                    fix_sort_idx += list(times_list_sorted[ind_with_same_time].argsort() + ind_with_same_time[0])
                else:
                    fix_sort_idx.append(sort_idx[ind])
                ind += counts

            times_list_sorted = times_list_sorted[fix_sort_idx]

            dt = []
            for i, t_hw in enumerate(zip(times_list_sorted[1::], times_list_sorted[:-1])):
                dt_tmp = t_hw[0] - t_hw[1]
                if dt_tmp < 0:  # HW timing was zeroing during the same packets sprinkler
                    n_wrap = 1 + (round(
                        (pc_time_list_sorted[i + 1] - pc_time_list_sorted[i]) * 1000) // max_stat_param_val)
                    dt_tmp = (n_wrap * max_stat_param_val) + t_hw[0] - t_hw[1]
                dt.append(dt_tmp)

            return dt

        def check_if_nan(lst):
            is_nan = pd.isnull(lst)
            if (not isinstance(is_nan, np.ndarray) and is_nan) or (isinstance(is_nan, np.ndarray) and any(is_nan)):
                return True
            return False

        if check_if_nan(self.gw_data['time_from_start']) or check_if_nan(self.gw_data['stat_param']):
            return None

        estimate_diff_time = estimate_diff_packet_time(self.gw_data['stat_param'], self.gw_data['time_from_start'])
        if estimate_diff_time is None:
            return None
        elif len(self) == 3:
            if triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=1):
                return None
            else:
                for ratio in [2, 3, 4]:
                    if triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=ratio):
                        estimate_diff_time = [min(estimate_diff_time[0], estimate_diff_time[1]),
                                              max(estimate_diff_time[0], estimate_diff_time[1]) / ratio]
                        break
                if triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=1.5):
                    estimate_diff_time = [min(estimate_diff_time[0], estimate_diff_time[1]) / 2,
                                          max(estimate_diff_time[0], estimate_diff_time[1]) / 3]

        return int(min(estimate_diff_time))

    def extract_packet_data_by_name(self, key):
        """
        extract data from all packet attribute according to the key name
        :param key: the name of the data (the key of the relevant dictionary)
        :type key: str
        :return: list of the data
        :rtype: list
        """

        if key in self.packet_data:
            data_attr = self.packet_data[key]
        elif hasattr(self, 'decoded_data') and key in self.decoded_data:
            data_attr = self.decoded_data[key]
        elif key in self.gw_data:
            data_attr = self.gw_data[key]
        elif key in self.custom_data:
            data_attr = self.custom_data[key]
        else:
            self.printing_and_logging('key:{} does not exist in packet structure')
            return None
        if isinstance(data_attr, np.ndarray):
            data_attr = data_attr.tolist()
        if not isinstance(data_attr, list):
            data_attr = [data_attr]
        return data_attr

    def filter_by_sprinkler_id(self, sprinkler_ids):
        """
        keep only specific sprinklers in a packet according to the sprinkler id
        :param sprinkler_ids:
        :type sprinkler_ids: list
        :return: filtered packet
        :rtype: Packet or DecryptedPacket
        """

        def filter_per_attr(data_attr):
            for key in data_attr.keys():
                if isinstance(data_attr[key], list) or \
                        (isinstance(data_attr[key], np.ndarray) and data_attr[key].size > 1):
                    filtered_data = [data_attr[key][i] for i in sprinkler_ids]
                    if isinstance(data_attr[key], np.ndarray):
                        data_attr[key] = np.array(filtered_data)
                    else:
                        data_attr[key] = filtered_data
            return data_attr

        filtered_packet = self.copy()
        filtered_packet.gw_data = filter_per_attr(filtered_packet.gw_data)
        filtered_packet.custom_data = filter_per_attr(filtered_packet.custom_data)

        return filtered_packet

    def add_custom_data(self, custom_data):
        for key in custom_data.keys():
            if isinstance(custom_data[key], list):
                if len(custom_data[key]) == self.__len__():
                    self.custom_data[key] = custom_data[key]
                else:
                    self.printing_and_logging('add_custom_data failed - '
                                              'the custom data is a list of a different length than'
                                              ' the number of packets')
            else:
                self.custom_data[key] = self.__len__() * [custom_data[key]]


if __name__ == '__main__':
    packet_1 = {'packet': '03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C404200486436AE8F',
                'is_valid_tag_packet': True, 'adv_address': '03B28DCD9920', 'group_id': 'FE0000', 'rssi': 54,
                'stat_param': 44687,
                'time_from_start': 1.528374, 'counter_tag': 1}
    packet_2 = {'packet': '03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C404200486436AE82',
                'is_valid_tag_packet': True, 'adv_address': '03B28DCD9920', 'group_id': 'FE0000', 'rssi': 60,
                'stat_param': 44688,
                'time_from_start': 2.528374, 'counter_tag': 2}

    p1 = Packet(packet_1)
    p2 = Packet(packet_2)

    print(p1.get_packet_string(0))
    print(p1.get_average_rssi())

    print(p1 == p2)
    print(p1.append_to_sprinkler(p2))

    print(len(p1))
    print(p1.get_average_rssi())

    p1_dict = p1.as_dict()
    p1_df = pd.DataFrame(data=p1_dict)
    print(p1_df)

    print('end')



#  """
#    Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import numpy as np
import pandas
import pandas as pd
import copy

from wiliot_core.packet_data.packet import Packet


class PacketList(list):
    def __init__(self, packet_obj_type=Packet, packet_list_obj_type=None, list_custom_data=None):
        """
        :param packet_obj_type:
        :param packet_list_obj_type:
        :param list_custom_data: the packet_list custom data as dictionary
        :type list_custom_data: dict
        """
        self.packet_list = np.array([], dtype=Packet)  # ndarray - contains packets
        self.raw_packet_map_list = {}  # dict - keys: raw_packet, values: index in self.packet_list
        self.is_df_changed = True
        self.packet_df = None
        self.packet_obj_type = packet_obj_type
        if packet_list_obj_type is None:
            self.packet_list_obj_type = PacketList
        else:
            self.packet_list_obj_type = packet_list_obj_type

        if list_custom_data is not None:
            self.list_custom_data = list_custom_data
        else:
            self.list_custom_data = {}

    def __add__(self, other_packet_list):
        """
        merge 2 PacketList object using '+' sign: packet_list1+packet_list2

        :type other_packet_list: PacketList
        :param other_packet_list:

        :return: merged packet_list, not mandatory
        """
        self.is_df_changed = True
        for packet in other_packet_list.packet_list:
            raw_packet = packet.get_packet()
            if raw_packet in self.raw_packet_map_list.keys():
                raw_packet_list_index = self.raw_packet_map_list[raw_packet]
                self.packet_list[raw_packet_list_index].append_to_sprinkler(packet)
            else:
                self.packet_list = np.append(self.packet_list, packet)
                raw_packet_list_index = self.packet_list.size - 1
                self.raw_packet_map_list[raw_packet] = raw_packet_list_index
            self.print_live_stream(packet)
        return self

    def __len__(self):
        """
        Total amount of packets (sprinklers calculated separately)
        """
        return self.size()

    def __iter__(self):
        self.n = 0
        return iter(self.packet_list)

    def __next__(self):
        if self.n <= len(self):
            return self.packet_list[self.n]
        else:
            raise StopIteration

    def __getitem__(self, key):
        return self.packet_list[key]

    def __setitem__(self, key, packet: Packet):
        # raise ValueError("Set item is not supported yet by PacketList type")
        self.is_df_changed = True
        if isinstance(packet, Packet):
            old_packet = self.packet_list[key]
            del self.raw_packet_map_list[old_packet.get_packet_string(gw_data=False, process_packet=False)]
            self.raw_packet_map_list[packet.get_packet_string(gw_data=False, process_packet=False)] = key
            self.packet_list[key] = packet
        else:
            raise TypeError("Can only set Packet type to PacketList")

    def pop(self, index=0):
        """

        :param index: index to pop out of the list
        :type index: int
        :return: the shorter packet list
        :rtype: PacketList
        """
        packet_list = self.copy()
        raw_packet = np.take(packet_list.packet_list, index).packet_data['raw_packet']
        packet_list.packet_list = np.delete(packet_list.packet_list, index)
        packet_list.raw_packet_map_list.pop(raw_packet, None)
        return packet_list

    def copy(self):
        if self.size() == 0:
            return copy.deepcopy(self)
        else:
            copy_packet_list = self.packet_list_obj_type()
            for p in self.packet_list:
                copy_packet_list.append(p.copy())
            return copy_packet_list

    def size(self):
        """
        Total amount of packets, sprinklers count individually
        """
        packet_list_size = 0
        for packet in self.packet_list:
            packet_list_size += len(packet)
        return packet_list_size

    def append(self, packet: Packet, ignore_sprinkler=False):
        """
        Adds single Packet to PacketList

        :type packet: Packet
        :param packet: packet to be added to packet_list
        :type ignore_sprinkler: Bool
        :param ignore_sprinkler: allow duplicates packets from different sprinkler

        :return: packet_list
        """
        if packet.is_valid_packet:
            self.is_df_changed = True
            raw_packet = packet.get_packet()
            if raw_packet not in self.raw_packet_map_list.keys():
                self.packet_list = np.append(self.packet_list, packet)
                raw_packet_list_index = self.packet_list.size - 1
                self.raw_packet_map_list[raw_packet] = raw_packet_list_index
            else:
                raw_packet_list_index = self.raw_packet_map_list[raw_packet]
                self.packet_list[raw_packet_list_index].append_to_sprinkler(packet)

    def get_sprinkler(self, packet):
        """
        @param packet: a packet data type. Needs to be appended before using this function.
        @return: A packet sprinkler of this packet.
        """
        raw_packet = packet.get_packet()
        if raw_packet in self.raw_packet_map_list.keys():
            raw_packet_list_index = self.raw_packet_map_list[raw_packet]
            return self.packet_list[raw_packet_list_index]
        else:
            return None

    def dump(self, packet_dict_list: list):
        """
        gets list of raw_packet or packet_dict and fill packet_list with data

        :type packet_dict_list: list
        :param packet_dict_list: gw list (get_packets), fill packet_list with data

        :return: bool status
        """
        if len(self) > 0:
            # packet_list not empty
            return False
        self.is_df_changed = True
        for packet_dict in packet_dict_list:
            packet = self.packet_obj_type(packet_dict)
            if packet.is_valid_packet:
                self.print_live_stream(packet)
                self.append(packet)
        return True

    def print_live_stream(self, packet):
        """
        for future use - implement output
        """
        # set parameters to filter view by
        pass

    def get_df(self, sprinkler_filter=False, add_sprinkler_info=False, is_overwrite=False, as_dict=False):
        """
        returns packet_list data as dataframe, BE CAREFULL WHEN CHANGING FUNCTIONALITY, this is a fundamental function

        :type sprinkler_filter: Bool
        :param sprinkler_filter: determine if to keep all occurrences of sprinkler per packet
        :type add_sprinkler_info: Bool
        :param add_sprinkler_info: adds 'num_packets_per_sprinkler','per','tbp' attributes to df
        :param is_overwrite: if True, get df overwrite columns with the same name
        :type is_overwrite: bool
        :param as_dict: if True, returns a dictionary instead of dataframe
        :type as_dict: bool
        :return: Dataframe
        """
        if not self.is_df_changed:
            sprinkler_info_calculated = 'packet_id' in self.packet_df.keys()
            if sprinkler_filter and sprinkler_info_calculated:
                return self.packet_df.drop_duplicates(subset='packet_id')
            elif (sprinkler_info_calculated and not is_overwrite) or not add_sprinkler_info:
                return self.packet_df

        all_packets_dict = []
        for packet_id, packet in enumerate(self.packet_list):
            sprinkler_dict = packet.as_dict()
            # add sprinkler info:
            if sprinkler_filter is True or add_sprinkler_info is True:
                tbp = packet.get_tbp()
                if tbp is None:
                    tbp = -1
                per = packet.get_per()

                sprinkler_keys = ['packet_id', 'sprinkler_counter', 'num_packets_per_sprinkler', 'tbp', 'per']
                sprinkler_values = [[packet_id] * len(packet), list(range(1, len(packet) + 1)),
                                    [len(packet)] * len(packet), [tbp] * len(packet), [per] * len(packet)]
                for k, v in zip(sprinkler_keys, sprinkler_values):
                    if k in sprinkler_dict.keys() and not is_overwrite:
                        continue
                    sprinkler_dict[k] = v
            # append packet dict:
            if isinstance(sprinkler_dict['raw_packet'], list) and len(sprinkler_dict['raw_packet']) > 1:
                for ind in range(len(sprinkler_dict['raw_packet'])):
                    all_packets_dict.append({k: v[ind] for k, v in sprinkler_dict.items()})
            else:
                for k, v in sprinkler_dict.items():
                    if isinstance(v, list):
                        sprinkler_dict[k] = v[0]
                    elif isinstance(v, np.ndarray):
                        sprinkler_dict[k] = v.item()
                all_packets_dict.append(sprinkler_dict)

        self.packet_df = pd.DataFrame(all_packets_dict)
        self.is_df_changed = False

        if as_dict:
            return all_packets_dict
        if sprinkler_filter:
            return self.packet_df.drop_duplicates(subset='packet_id')
        else:
            return self.packet_df

    def packet_df_to_sprinkler_df(self, packet_df):
        """
        gets packet_df and returns sprinkler df
        :type packet_df: DataFrame
        :param packet_df: df to convert
        :return: Dataframe - sprinkler_df
        """
        packet_list = self.import_packet_df(packet_df=packet_df)
        sprinkler_df = packet_list.get_df(sprinkler_filter=True, add_sprinkler_info=True)

        return sprinkler_df

    def sort_df_by(self, column='stat_param'):
        """
        returns dataframe sorted by column
        :type column: str
        :param column: the column to filter by

        :return: Dataframe
        """
        packet_df = self.get_df()
        packet_df_sorted = packet_df.sort_values(column)
        return packet_df_sorted

    def filter_packet_by(self, packet_data_key='adv_address', values='', filter_out=False):
        """
        filter packet_list by adv_address or any other data key
        :type packet_data_key: str
        :param packet_data_key:
        :type values: str or list of strings
        :param values: e.g. adv_address to search
        :type filter_out: bool
        :param filter_out: if True return all packets that does NOT have packet_data_key values

        :return: filtered PacketList
        """
        packet_list = self.packet_list_obj_type(list_custom_data=self.list_custom_data)
        for index, packet in enumerate(self.packet_list):
            filter_data = packet.extract_packet_data_by_name(packet_data_key)
            if filter_data is None:
                print('PacketList->filter_packet_by: packet_data_key does not exist')
                return None

            if not isinstance(values, list):
                values = [values]

            if len(filter_data) > 1:  # filter by sprinklers
                filtered_sprinkler_id = [i for i, f_d in enumerate(filter_data) if (not filter_out and f_d in values) or
                                         (filter_out and f_d not in values)]
                if len(filtered_sprinkler_id):
                    packet_list.append(self.packet_list[index].filter_by_sprinkler_id(filtered_sprinkler_id))
            else:  # filter by packet
                if (not filter_out and filter_data[0] in values) or (filter_out and filter_data[0] not in values):
                    packet_list.append(self.packet_list[index])

        return packet_list

    def filter_df_by(self, packet_df=None, column='adv_address', values='', values_range=[]):
        """
        filter dataframe by value or (exclusive) by value range
        :type packet_df: DataFrame
        :param packet_df: gets get_df dataframe (packet_df only)
        :type column: str
        :param column: column to filter by
        :type values: str or int or list of strings
        :param values: value to search
        :type values_range: 2 elements list of int
        :param values_range: range to find

        :return: filtered Dataframe
        """
        if packet_df is None:
            packet_df = self.get_df()
        if values_range:
            start = values_range[0]
            end = values_range[1]
            packet_df_filtered = packet_df.loc[(packet_df[column] > start) & (packet_df[column] <= end)]
        else:
            if isinstance(values, list):
                packet_df_filtered = packet_df.loc[packet_df[column].isin(values)]
            else:
                packet_df_filtered = packet_df.loc[packet_df[column] == values]
        return packet_df_filtered

    def get_avg_rssi(self):
        """
        return packet list average rssi (4 decimal points accuracy)
        :return: average rssi for tag
        """
        df = self.get_df()
        if df.empty:
            return None
        avg_rssi = round(df['rssi'].mean(), 4)
        return avg_rssi

    def get_avg_tbp(self, ignore_outliers=False):
        """
        return packet list average tbp (4 decimal points accuracy)
        :param ignore_outliers: reject data outside the 2 std area
        :type ignore_outliers: bool

        :return: average tbp for tag
        """

        def reject_outliers(data, m=2):
            return data[abs(data - np.mean(data)) < m * np.std(data)]

        tbp_list = np.array([])
        for packet in self.packet_list:
            packet_tbp = packet.get_tbp()
            if packet_tbp is not None:
                tbp_list = np.append(tbp_list, packet_tbp)
        if len(tbp_list) == 0:
            return None
        avg_tbp = round(tbp_list.mean(), 4)
        if ignore_outliers:
            avg_tbp = round(reject_outliers(tbp_list).mean(), 4)
        return avg_tbp

    def to_csv(self, path, append=False, export_packet_id=True, columns=None, add_sprinkler_info=False):
        """
        export entire PacketList to csv

        :type path: str
        :param path: path to save csv
        :type append: bool
        :param append: to append the df to an existing csv file

        :return: bool - export status
        """

        return self.export_packet_df(packet_df=self.get_df(add_sprinkler_info=add_sprinkler_info), path=path,
                                     append=append,
                                     export_packet_id=export_packet_id, columns=columns)

    def export_packet_df(self, packet_df, path, append=False, export_packet_id=True, columns=None):
        """
        export given dataframe to csv

        :type packet_df: Dataframe
        :param packet_df: filtered dataframe to save as csv
        :type path: str
        :param path: path to save csv
        :type append: bool
        :param append: to append the df to an existing csv file

        :return: bool - export status
        """
        try:
            # this call of 'to_csv' is a generic pandas function
            if not export_packet_id and 'packet_id' in packet_df:
                packet_df.drop('packet_id', axis=1, inplace=True)

            if append:
                packet_df.to_csv(path, mode='a', index=False, header=False, columns=columns)
            else:
                packet_df.to_csv(path, index=False, columns=columns)
            return True
        except Exception as e:

            raise Exception('problem during export packet df {}'.format(e))

    def import_packet_df(self, path=None, packet_df=None, custom_data_attr=None, import_all=False,
                         list_custom_data=None, verbose=True):
        """
        import from a csv of dataframe
        :type path: str
        :param path: the message or part of the message that needed to be read
        :type packet_df: DataFrame
        :param packet_df: gets get_df dataframe (packet_df only)
        :type custom_data_attr: list
        :param custom_data_attr: list of all the custom data names
        :param list_custom_data: the packet_list custom data as dictionary
        :type list_custom_data: dict
        :return: PacketList
        """
        non_packets_rows = []
        list_custom_data = list_custom_data if list_custom_data is not None else self.list_custom_data
        import_packet_list = self.packet_list_obj_type(list_custom_data=list_custom_data)
        if path is not None:
            import_packets_df = pd.read_csv(path)
        else:
            import_packets_df = packet_df

        for index, row in import_packets_df.iterrows():
            try:
                reconstructed_packet = None
                if 'raw_packet' in row.keys() and 'gw_packet' in row.keys():
                    raw_packet = row['raw_packet']
                    gw_data = row['gw_packet']
                    if not pd.isnull(raw_packet) and not pd.isnull(gw_data):
                        reconstructed_packet = str(raw_packet) + str(gw_data)
                elif 'encryptedPacket' in row.keys():
                    reconstructed_packet = row['encryptedPacket']
                elif 'decrypted_full_packet' in row.keys():
                    reconstructed_packet = row['decrypted_full_packet']
                elif 'encrypted_packet' in row.keys():
                    reconstructed_packet = row['encrypted_packet']
                else:
                    non_packets_rows.append(index)
                    continue
                if reconstructed_packet == '' or pd.isnull(reconstructed_packet):
                    non_packets_rows.append(index)
                    continue
            except Exception as e:
                non_packets_rows.append(index)
                continue

            if 'time_from_start' in row.keys():
                time_from_start = row['time_from_start']
            elif 'time' in row.keys():
                time_from_start = row['time']
            else:
                time_from_start = float('nan')
            if time_from_start == '' or time_from_start is None:
                time_from_start = float('nan')

            # adding custom data
            custom_data = {}
            if import_all:
                try:
                    p_tmp = self.packet_obj_type(reconstructed_packet, float(time_from_start))
                    static_keys_list = p_tmp.get_packet_data_names()
                    for k in row.keys():
                        if k not in static_keys_list and k != 'packet_id':
                            custom_data[k] = row[k]
                except Exception as e:
                    print('exception raised during extracting custom data in import df: {}'.format(e))
            elif custom_data_attr is not None:
                for k in custom_data_attr:
                    if k in row.keys():
                        custom_data[k] = row[k]
            try:
                p = self.packet_obj_type(reconstructed_packet, float(time_from_start), custom_data=custom_data)
                import_packet_list.append(p)
            except Exception as e:
                print('exception raised during import df: {}'.format(e))

        if len(non_packets_rows) and verbose:
            print('the following {} rows could not be imported to the packet list:{}'.format(len(non_packets_rows),
                                                                                             non_packets_rows))
        return import_packet_list

    def get_statistics(self):
        """
        Calculates statistics of self.
        @return dictionary with predefined statistics of the packetList.
        """
        return self.get_df_statistics()

    def get_group_statistics(self, group_by_col='adv_address'):
        """
        Calculates statistics of self, grouped by some value.
        @return dictionary of items at the group, Dictionary values are dictionaries of statistics.
        """
        packet_df = self.get_df()
        # if not defines - get all values:

        groups_id_list = packet_df[group_by_col].unique()

        group_statistics = {}
        for group_id in groups_id_list:
            packet_df_filtered = packet_df.loc[packet_df[group_by_col] == group_id]
            group_statistics[group_id] = self.get_df_statistics(packet_df=packet_df_filtered)
            group_statistics[group_id][group_by_col] = group_id
        return group_statistics

    def group_by(self, group_by_col='adv_address'):
        """
        Calculates statistics of self, grouped by some value.
        @return dictionary of items at the group, Dictionary values are dictionaries of statistics.
        """
        packet_df = self.get_df(sprinkler_filter=True)
        # if not defines - get all values:

        groups_id_list = packet_df[group_by_col].unique()

        group = {}
        for group_id in groups_id_list:
            group[group_id] = self.filter_packet_by(group_by_col, group_id)
            # group[group_id][group_by_col] = group_id
        return group

    def get_num_packets(self):
        """
        Total amount of packets, sprinklers packets count separately
        """
        return len(self.packet_list)

    def get_df_statistics(self, packet_df=None):
        """
        Calculates statistics from packetList DF.
        @param packet_df - dataframe generated by packetList
        @return dictionary with predefined statistics of the packetList.
        """
        statistics = {}
        if (packet_df is not None and packet_df.shape[0] == 0) or (packet_df is None and self.size() == 0):
            statistics = {'num_packets': 0, 'num_cycles': 0}
            return statistics

        # sprinkler_df = self.get_df(sprinkler_filter=True)
        if packet_df is None:
            packet_df = self.get_df(add_sprinkler_info=True)
        if 'sprinkler_counter' in packet_df.keys():
            sprinkler_df = packet_df[packet_df['sprinkler_counter'] == 1]
        else:
            sprinkler_df = self.packet_df_to_sprinkler_df(packet_df)

        statistics['num_packets'] = packet_df.shape[0]
        statistics['num_cycles'] = sprinkler_df.shape[0]

        sprinkler_counter_arr = pandas.to_numeric(sprinkler_df['num_packets_per_sprinkler'], errors='coerce')
        statistics['sprinkler_counter_mean'] = np.mean(sprinkler_counter_arr)
        statistics['sprinkler_counter_std'] = np.std(sprinkler_counter_arr)
        statistics['sprinkler_counter_min'] = np.min(sprinkler_counter_arr)
        statistics['sprinkler_counter_max'] = np.max(sprinkler_counter_arr)

        tbp_arr = pandas.to_numeric(sprinkler_df.loc[sprinkler_df['tbp'] > 0]['tbp'], errors='coerce')
        if len(tbp_arr) == 0:
            tbp_arr = [-1]
        statistics['tbp_mean'] = np.mean(tbp_arr)
        statistics['tbp_std'] = np.std(tbp_arr)
        statistics['tbp_min'] = np.min(tbp_arr)
        statistics['tbp_max'] = np.max(tbp_arr)
        statistics['tbp_num_vals'] = np.size(tbp_arr)

        per_arr = pandas.to_numeric(sprinkler_df['per'], errors='coerce')
        statistics['per_mean'] = np.mean(per_arr)
        statistics['per_std'] = np.std(per_arr)

        rssi_arr = pandas.to_numeric(packet_df['rssi'], errors='coerce')
        statistics['rssi_mean'] = np.mean(rssi_arr)
        statistics['rssi_std'] = np.std(rssi_arr)
        statistics['rssi_min'] = np.min(rssi_arr)
        statistics['rssi_max'] = np.max(rssi_arr)

        time_arr = pandas.to_numeric(packet_df['time_from_start'], errors='coerce')
        # statistics['start_time'] = np.min(time_arr)
        statistics['ttfp'] = np.min(time_arr)
        statistics['end_time'] = np.max(time_arr)
        statistics['duration'] = statistics['end_time'] - statistics['ttfp']
        if statistics['duration'] > 0:
            statistics['rx_rate_normalized'] = statistics['num_packets'] / statistics['duration']
            statistics['rx_rate'] = statistics['num_packets'] / statistics['end_time']
        else:
            statistics['rx_rate'] = 0
            statistics['rx_rate_normalized'] = 0
        return statistics

    def to_tag_list(self):
        # import here to avoid infinte loop (packet_list calls tag_list and tags_list calls packet_list)
        from wiliot_core.packet_data.tag_collection import TagCollection
        tag_list = TagCollection()
        for packet in self.packet_list:
            tag_list.append(packet)
        return tag_list


if __name__ == '__main__':
    import os

    try:
        df_path = os.path.abspath(
            '../../internal/data_samples_for_tests/offline_tester@packets_data.csv')

        new_list = PacketList()
        new_list.append(Packet(
            'process_packet("04A3E93988271E16AFFD0200004E9D46293BEED581709F1214C88F6F3978E2136EB48E413533807C")'))
        new_list.get_df()

        new_list = new_list.import_packet_df(path=df_path, import_all=True)
        new_list_df = new_list.get_df(sprinkler_filter=True, add_sprinkler_info=True, is_overwrite=True)
        new_list.to_csv(
            path=os.path.abspath(
                '../../internal/data_samples_for_tests/offline_tester@packets_data_decrypted.csv'))
        new_list_short = PacketList()
        new_list_short = new_list_short.import_packet_df(path=df_path, custom_data_attr=['common_run_name',
                                                                                         'external_id'])
        new_list_short_df = new_list_short.get_df(add_sprinkler_info=True)

    except Exception as e:
        print(e)

    print('done')



"""
Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import serial
import time
from multiprocessing import Process, Queue, Event
from queue import Empty
from enum import Enum
import datetime
import logging
import os
from wiliot_core.utils.utils import WiliotDir

packet_bytes = 100  # number of char
packet_prefix_str = 'process_packet'


class SerialProcessState(Enum):
    CONNECT = 'connect'
    WRITE = 'write'
    READ = 'read'
    DISCONNECT = 'disconnect'
    RESET = 'reset'
    STOP = 'stop'
    IDLE = 'idle'


class SerialProcess:
    """
    connect - connect to serial, return if connected (connected event)
    write - write command to serial, return the gw respond (add to rsp q)
    read - read specific msg from serial, return the gw respond (add to rsp/packet q)
    disconnect - close serial port, return if connected (connected event)
    reset - reset serial buffer
    check - number of bytes waiting in buffer (n_bytes Value
    stop - stop the process
    """
    def __init__(self, cmd_q, gw_rsp_q, packets_q, change_connection_status, error_event, log_dir_path=None):
        """

        :param cmd_q: queue of commands to the listener including cmd = the state in the state machine,
                      and data if needed
        :type cmd_q: Queue
        :param gw_rsp_q: the listener put into this queue the gw responds
        :type gw_rsp_q: Queue
        :param packets_q: the listener put into this queue the packets received from gw
        :type packets_q: Queue
        :param change_connection_status: listener set this event if there was a change in the connection status
        :type change_connection_status: Event
        :param error_event: listener set this event if an error occurs during run
        :type error_event: Event
        :param log_dir_path: the logging directory
        :type log_dir_path: str
        """
        # serial port variables:
        self._comPortObj = None
        self.write_wait_time = 0.001  # [sec]
        self.cmd_q = cmd_q
        self.gw_rsp_q = gw_rsp_q
        self.packets_q = packets_q
        self.change_connection_status = change_connection_status
        self.error_event = error_event
        self.connected = False
        self.gw_start_time = datetime.datetime.now()
        self.buf = b''

        # set logging:
        self.logger = logging.getLogger('WiliotListener')
        self.set_logger(log_dir_path)
        # start run:
        self.run_sm()

    def set_logger(self, logger_dir_path):
        formatter = logging.Formatter('%(asctime)s,%(msecs)d %(name)s %(levelname)s: %(message)s', '%Y-%m-%d %H:%M:%S')
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.INFO)
        stream_handler.setFormatter(formatter)
        if logger_dir_path is None:
            wiliot_dir = WiliotDir()
            logger_dir_path = os.path.join(wiliot_dir.get_wiliot_root_app_dir(), 'wiliot_continuous_listener')
        if not os.path.isdir(logger_dir_path):
            os.mkdir(logger_dir_path)
        logger_name = 'listener_log_{}.log'.format(datetime.datetime.now().strftime('%Y%m%d_%H%M%S'))
        logger_path = os.path.join(logger_dir_path, logger_name)
        file_handler = logging.FileHandler(logger_path, mode='a')
        file_formatter = logging.Formatter('%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s', '%H:%M:%S')
        file_handler.setFormatter(file_formatter)
        self.logger.addHandler(file_handler)
        self.logger.addHandler(stream_handler)
        self.logger.setLevel(logging.DEBUG)

    def run_sm(self):
        while True:
            try:
                cmd = self.cmd_q.get(timeout=0)
                state = cmd['cmd']
            except Empty:
                cmd = None
                state = SerialProcessState.IDLE

            if state == SerialProcessState.CONNECT:
                self.connect(port=cmd['data']['port'], baud=cmd['data']['baud'])
            elif state == SerialProcessState.WRITE:
                self.write(gw_cmd=cmd['data']['gw_cmd'])
                self.read()
            elif state == SerialProcessState.READ:
                self.gw_start_time = datetime.datetime.now()
                self.read()
            elif state == SerialProcessState.DISCONNECT:
                self.disconnect()
            elif state == SerialProcessState.RESET:
                self.reset()
                self.gw_start_time = datetime.datetime.now()
            elif state == SerialProcessState.STOP:
                self.disconnect()
                break
            elif state == SerialProcessState.IDLE:
                if self.connected:
                    self.read()
                else:
                    time.sleep(0.1)
                    continue
        self.logger.log(logging.INFO, 'serial process state machine is done')

    def connect(self, port, baud):
        if self.connected:
            self.change_connection_status.set()
        else:
            try:
                self._comPortObj = serial.Serial(port, baud, timeout=0.001)
                for i in range(5):
                    if self._comPortObj.isOpen():
                        break
                    time.sleep(0.1)
                    self._comPortObj = serial.Serial(port, baud, timeout=0.001)
                if self._comPortObj.isOpen():
                    self.change_connection_status.set()
                    self.connected = True
                    self.logger.log(logging.INFO, 'Connected to port: {}'.format(port))
            except Exception as e:
                self.logger.log(logging.WARNING, 'Exception during connection to port {}: {}'.format(port, e))

    def disconnect(self):
        if self.connected:
            try:
                self._comPortObj.close()
                self.change_connection_status.set()
                self.connected = False
            except Exception as e:
                self.logger.log(logging.WARNING, 'Exception during disconnect from serial: {}'.format(e))

    def write(self, gw_cmd):
        if self.connected:
            if isinstance(gw_cmd, str):
                gw_cmd = gw_cmd.encode()
            if gw_cmd != b'':
                if len(gw_cmd) >= 2:  # check last characters
                    if gw_cmd[-2:] != b'\r\n':
                        gw_cmd += b'\r\n'
                else:
                    gw_cmd += b'\r\n'
                if gw_cmd[0:1] != b'!':  # check first character
                    gw_cmd = b'!' + gw_cmd

                try:
                    self._comPortObj.write(gw_cmd)
                    self.logger.log(logging.DEBUG, "sent msg {}".format(gw_cmd))
                except Exception as e:
                    self.logger.log(logging.WARNING, "failed to send the command to GW (check your physical "
                                                     "connection:\n{}".format(e))

    def reset(self):
        if self.connected:
            try:
                self._comPortObj.flush()  # wait until all data is written to the comport
                if self._comPortObj.in_waiting:
                    self.logger.log(logging.WARNING, 'flush input buffer with {} bytes waiting'.format(
                        self._comPortObj.in_waiting))
                    self._comPortObj.reset_input_buffer()
            except Exception as e:
                self.logger.log(logging.WARNING, 'Exception during reset port: {}'.format(e))

    def read(self):
        if self.connected:
            try:
                data = self._comPortObj.readline()
                self.buf += data
                if b'\n' in data:
                    msg = self.buf.decode().strip(' \t\n\r')
                    packet_time = datetime.datetime.now() - self.gw_start_time
                    msg_dict = {'time': packet_time.total_seconds(), 'raw': msg}
                    self.logger.log(logging.DEBUG, "got msg {}".format(msg_dict))
                    if msg.startswith(packet_prefix_str):
                        # got a packet
                        if self.packets_q.full():
                            dummy = self.packets_q.get()
                            self.logger.log(logging.WARNING, "packets_q is full, dropping {}".format(dummy))
                        self.packets_q.put(msg_dict)
                    else:
                        # got a responds
                        if self.gw_rsp_q.full():
                            dummy = self.gw_rsp_q.get()
                            self.logger.log(logging.WARNING, "gw_rsp_q is full, dropping {}".format(dummy))
                        self.gw_rsp_q.put(msg_dict)
                    self.buf = b''

                # check buffer status:
                if self._comPortObj.in_waiting > 2 * packet_bytes:
                    self.logger.log(logging.WARNING, 'buffer contains {} bytes'.format(self._comPortObj.in_waiting))
            except serial.SerialException as e:
                self.logger.log(logging.WARNING, ' Serial Exception during reading: {}'.format(e))
                if 'PermissionError' in str(e):
                    self.logger.log(logging.WARNING, 'disconnecting')
                    self.disconnect()
            except Exception as e:
                self.error_event.set()
                self.logger.log(logging.WARNING, ' problem during reading: {}\nreset buffer'.format(e))
                self.reset()


if __name__ == '__main__':
    commands = Queue(maxsize=100)
    gw_rsp = Queue(maxsize=100)
    packets = Queue(maxsize=100)
    connection_status = Event()
    event_error = Event()

    p_thread = Process(target=SerialProcess, args=(commands, gw_rsp, packets, connection_status,event_error))
    p_thread.start()
    time.sleep(1)
    commands.put({'cmd': SerialProcessState.CONNECT, 'data': {'port': 'COM3', 'baud': 921600}})
    connection_status.wait(3)
    if connection_status.is_set():
        print('connected')
        commands.put({'cmd': SerialProcessState.WRITE, 'data': {'gw_cmd': '!version'}})
        try:
            version_msg = gw_rsp.get(timeout=1)
            print('got rsp: {}'.format(version_msg))
        except Empty:
            print('empty')
    else:
        print('problem during connection')



#  """
#    Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import os
import csv
from appdirs import user_data_dir
import logging
import json
import pathlib
try:
    import PySimpleGUI as SimGUI
except Exception as e:
    print('could not import PySimpleGUI due to {}'.format(e))


class WiliotDir:
    def __init__(self) -> None:
        self.local_appdata_dir = ''
        self.wiliot_root_path = ''
        self.common_dir_path = ''
        self.config_dir_path = ''
        self.user_config_path = ''
        self.tester_subdirectories = ['results', 'logs', 'configs']
        
        self.set_dir()
        self.create_dir(self.local_appdata_dir)
        self.create_dir(self.wiliot_root_path)
        self.create_dir(self.common_dir_path)
        self.create_dir(self.config_dir_path)
    
    def set_dir(self):
        try:
            if 'WILIOT_APP_ROOT_PATH' in os.environ.keys():
                print(os.environ['WILIOT_APP_ROOT_PATH'])
                self.wiliot_root_path = os.environ['WILIOT_APP_ROOT_PATH']
            else:
                self.local_appdata_dir = user_data_dir()
                self.wiliot_root_path = os.path.abspath(os.path.join(self.local_appdata_dir, 'wiliot'))
            
            self.common_dir_path = os.path.abspath(os.path.join(self.wiliot_root_path, 'common'))
            self.config_dir_path = os.path.abspath(os.path.join(self.common_dir_path, 'configs'))
            self.user_config_path = os.path.abspath(os.path.join(self.config_dir_path, 'user_configs.json'))
        
        except Exception as e:
            logging.warning('Error loading environment or getting in from OS, supporting Windows, Linux and MacOS '
                            '({})'.format(e))
    
    @staticmethod
    def create_dir(path):
        if not os.path.exists(path):
            os.makedirs(path)
    
    def create_tester_dir(self, tester_name):
        tester_path = self.get_tester_dir(tester_name)
        self.create_dir(tester_path)
        
        for subdir in self.tester_subdirectories:
            self.create_dir(tester_path + '/' + subdir)
    
    def get_tester_dir(self, tester_name):
        wiliot_path = self.wiliot_root_path
        tester_path = os.path.abspath(os.path.join(wiliot_path, tester_name))
        return tester_path
    
    def get_dir(self):
        return self.wiliot_root_path, self.common_dir_path, self.config_dir_path, self.user_config_path
    
    def get_wiliot_root_app_dir(self):
        return self.wiliot_root_path
    
    def get_common_dir(self):
        return self.common_dir_path
    
    def get_config_dir(self):
        return self.config_dir_path
    
    def get_user_config_file(self, client_type=None):
        if client_type is None:
            return self.user_config_path
        return self.user_config_path.replace('.json', f'_{client_type}.json')


def open_json(folder_path, file_path, default_values=None):
    """
    opens config json
    :type folder_path: string
    :param folder_path: the folder path which contains the desired file
    :type file_path: string
    :param file_path: the file path which contains the json
            (including the folder [file_path = folder_path+"json_file.json"])
    :type default_values: dictionary
    :param default_values: default values for the case of empty json
    :return: the desired json object
    """
    if not os.path.exists(folder_path):
        pathlib.Path(folder_path).mkdir(parents=True, exist_ok=True)
    
    file_exists = os.path.isfile(file_path)
    if not file_exists or os.stat(file_path).st_size == 0:
        # save the default values to json
        with open(file_path, "w") as out_file:
            json.dump(default_values, out_file)
        
        return json.load(open(file_path, "rb"))
    else:
        with open(file_path) as f:
            json_content = f.read()
        if len(json_content) == 0:
            with open(file_path, "w") as out_file:
                json.dump(default_values, out_file)
            json_content = json.load(open(file_path, "rb"))
        else:
            json_content = json.loads(json_content)
        return json_content


def credentials_gui(client_type=None):
    """
    open GUI for getting api_key from user
    :param client_type: the cloud client type. relevant if you have different api key for several client,
                       which will be save in different yser_config files
    :type client_type: str
    :return: the user input, dict with api_key, owner_id and env
    :rtype: dict
    """
    if client_type is None:
        client_type = 'DEFAULT'
    layout = [
        [SimGUI.Text('Please enter your Wiliot API key')],
        [SimGUI.Text(f'Client Type: {client_type}')],
        [SimGUI.Text('API key:'),
         SimGUI.InputText('', key='api_key')],
        [SimGUI.Text('owner id (optional):'),
         SimGUI.InputText('', key='owner_id')],
        [SimGUI.Text('environment (optional):'),
         SimGUI.InputCombo(('prod', 'non-prod'), default_value='prod', key='env')],
        [SimGUI.Submit()]]
    
    window = SimGUI.Window('User Credentials', layout)
    while True:
        event, values = window.read()
        if event == 'Submit':
            break
        elif event is None:
            print('User exited the program')
            window.close()
            break
    
    window.close()
    return values


def check_user_config_is_ok(owner_id=None, env=None, client_type=None):
    """
    check user credentials
    :param owner_id: use the key of the specified owner id
    :type owner_id: str or None
    :param env: use the key of the specified environment
    :type env: str or None
    :param client_type: if specified the credential will be read/write from the same use config file,
                        but with the specified suffix
    :type client_type: str or None
    :return: user_config_file_path - the file path,
             api_key -the key,
             is_success - True if the credentials we entered correctly
    :rtype: tuple
    """
    # Create wiliot appdata directory if not exists:
    is_success = True
    auth_gui_is_needed = False
    api_key = None
    cfg_data = []
    env_dirs = WiliotDir()
    config_dir_path = env_dirs.get_config_dir()
    user_config_file_path = env_dirs.get_user_config_file(client_type)
    if env is not None:
        env = env.lower()
    if env == '':
        env = 'prod'
    if env is not None and ('test' in env or 'dev' in env):
        env = 'non-prod'
    
    if not os.path.isdir(config_dir_path):
        pathlib.Path(config_dir_path).mkdir(parents=True, exist_ok=True)

    if os.path.exists(user_config_file_path):
        cfg_data = open_json(folder_path=config_dir_path, file_path=user_config_file_path)
        try:
            api_key_list = cfg_data['api_key']
            if not isinstance(api_key_list, list) or api_key_list == []:
                print('api key is missing. Please enter it manually')
                is_success = False
                auth_gui_is_needed = True
            elif (owner_id is not None and owner_id not in cfg_data['owner_id']) or \
                    (env is not None and env not in cfg_data['env']):
                print('api key does not match the request owner id or env. Please enter it manually')
                auth_gui_is_needed = True
            else:
                api_key = [k for i, k in enumerate(api_key_list)
                           if (owner_id is None or owner_id == cfg_data['owner_id'][i]) and
                           (env is None or env == cfg_data['env'][i])]
                if api_key == []:
                    print('could not find the api key for the specified environment and owner id')
                    auth_gui_is_needed = True
                api_key = api_key[0]
        except Exception as e:
            auth_gui_is_needed = True
            print("Config file is not readable at path {}. Exception {}\n Please enter new credentials".format(
                user_config_file_path, e))
    else:
        print("Config file user_configs.json doesn't exist at {}\n".format(config_dir_path))
        auth_gui_is_needed = True
    
    while auth_gui_is_needed:
        auth_gui_is_needed = False
        values = credentials_gui(client_type)
        api_key = values['api_key']
        if api_key == '':
            print('api key is missing. Please try again\n')
            auth_gui_is_needed = True
        elif api_key is None:
            is_success = False
        else:
            with open(user_config_file_path, 'w') as cfg:
                if isinstance(cfg_data, dict):
                    for k in values.keys():
                        if k in cfg_data.keys() and isinstance(cfg_data[k], list):
                            cfg_data[k].append(values[k])
                        else:
                            cfg_data[k] = [values[k]]
                    json.dump(cfg_data, cfg)
                else:
                    json.dump({'api_key': [api_key], 'owner_id': [values['owner_id']], 'env': [values['env']]}, cfg)
            is_success = True

    return user_config_file_path, api_key, is_success


def csv_to_dict(path=None):
    """
    convert csv to dictionary (arranged by columns)
    :param path: the csv path
    :type path: str
    :return: the data as dictionary
    :rtype: dict
    """
    if path:
        with open(path, 'r') as f:
            reader = csv.DictReader(f, delimiter=',')
            col_names = reader.fieldnames
            data_out = {col: [] for col in col_names}
            try:
                for row in reader:
                    for col in col_names:
                        data_out[col].append(row[col])
            except Exception as e:
                print("couldn't load csv due to {}".format(e))
        return data_out
    else:
        print('please provide a path')
        return None


if __name__ == '__main__':
    user_config_file_path, api_key, is_success = check_user_config_is_ok(env='non-prod')
    pass



#  """
#    Copyright (c) 2016- 2023, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

# package functions -
# gateway
from wiliot_core.local_gateway.local_gateway_core import *

# packet data
from wiliot_core.packet_data.packet import *
from wiliot_core.packet_data.packet_list import *
from wiliot_core.packet_data.tag_collection import *

# utils
from wiliot_core.utils.utils import *

# package internal functions
try:
    from wiliot_core.packet_data.extended.decrypted_packet import *
    from wiliot_core.packet_data.extended.decrypted_packet_list import *
    from wiliot_core.packet_data.extended.decrypted_tag_collection import *
except Exception as e:
    pass
